@@ -30,6 +30,7 @@
 #include "backend.h"
 #include "utils/debug.h"
 #include "utils/libvlc.h"
+#include "utils/vstring.h"
 
 QT_BEGIN_NAMESPACE
 
@@ -42,9 +43,7 @@ namespace VLC
  * Device Info
  */
 
-DeviceInfo::DeviceInfo(const QByteArray &name,
-                       const QString &description,
-                       bool isAdvanced)
+DeviceInfo::DeviceInfo(const QString &name, bool isAdvanced)
 {
     // Get an id
     static int counter = 0;
@@ -52,9 +51,12 @@ DeviceInfo::DeviceInfo(const QByteArray &name,
 
     // Get name and description for the device
     m_name = name;
-    m_description = description;
     m_isAdvanced = isAdvanced;
     m_capabilities = None;
+
+    // A default device should never be advanced
+    if (name.startsWith(QLatin1String("default"), Qt::CaseInsensitive))
+        m_isAdvanced = false;
 }
 
 int DeviceInfo::id() const
@@ -62,7 +64,7 @@ int DeviceInfo::id() const
     return m_id;
 }
 
-const QByteArray& DeviceInfo::name() const
+const QString& DeviceInfo::name() const
 {
     return m_name;
 }
@@ -89,6 +91,8 @@ const DeviceAccessList& DeviceInfo::accessList() const
 
 void DeviceInfo::addAccess(const DeviceAccess& access)
 {
+    if (m_accessList.isEmpty())
+        m_description = access.first + ": " + access.second;
     m_accessList.append(access);
 }
 
@@ -117,17 +121,6 @@ DeviceManager::DeviceManager(Backend *parent)
 
 DeviceManager::~DeviceManager()
 {
-    m_devices.clear();
-}
-
-int DeviceManager::deviceId(const QByteArray &name) const
-{
-    foreach (const DeviceInfo &device, m_devices) {
-        if (device.name() == name)
-            return device.id();
-    }
-
-    return -1;
 }
 
 QList<int> DeviceManager::deviceIds(ObjectDescriptionType type)
@@ -165,6 +158,7 @@ QHash<QByteArray, QVariant> DeviceManager::deviceProperties(int id)
             properties.insert("description", device.description());
             properties.insert("isAdvanced", device.isAdvanced());
             properties.insert("deviceAccessList", QVariant::fromValue<Phonon::DeviceAccessList>(device.accessList()));
+            properties.insert("discovererIcon", "vlc");
 
             if (device.capabilities() & DeviceInfo::AudioOutput) {
                 properties.insert("icon", QLatin1String("audio-card"));
@@ -186,7 +180,7 @@ QHash<QByteArray, QVariant> DeviceManager::deviceProperties(int id)
     return properties;
 }
 
-const DeviceInfo *DeviceManager::device(int id)
+const DeviceInfo *DeviceManager::device(int id) const
 {
     for (int i = 0; i < m_devices.size(); i ++) {
         if (m_devices[i].id() == id)
@@ -200,15 +194,11 @@ static QList<QByteArray> vlcAudioOutBackends()
 {
     QList<QByteArray> ret;
 
-    libvlc_audio_output_t *firstAudioOut = libvlc_audio_output_list_get(libvlc);
-    if (!firstAudioOut) {
-        error() << "libVLC:" << LibVLC::errorMessage();
-        return ret;
+    VLC_FOREACH_LIST(audio_output, aout) {
+        QByteArray name(aout->psz_name);
+        if (!ret.contains(name))
+            ret.append(name);
     }
-    for (libvlc_audio_output_t *audioOut = firstAudioOut; audioOut; audioOut = audioOut->p_next) {
-        ret.append(QByteArray(audioOut->psz_name));
-    }
-    libvlc_audio_output_list_release(firstAudioOut);
 
     return ret;
 }
@@ -217,10 +207,6 @@ void DeviceManager::updateDeviceList()
 {
     QList<DeviceInfo> newDeviceList;
 
-    DeviceInfo defaultAudioOutputDevice("default");
-    defaultAudioOutputDevice.setCapabilities(DeviceInfo::AudioOutput);
-    newDeviceList.append(defaultAudioOutputDevice);
-
     if (!LibVLC::self || !libvlc)
         return;
 
@@ -230,8 +216,10 @@ void DeviceManager::updateDeviceList()
     PulseSupport *pulse = PulseSupport::getInstance();
     if (pulse && pulse->isActive()) {
         if (audioOutBackends.contains("pulse")) {
-            defaultAudioOutputDevice.setAdvanced(false);
+            DeviceInfo defaultAudioOutputDevice(tr("Default"), false);
+            defaultAudioOutputDevice.setCapabilities(DeviceInfo::AudioOutput);
             defaultAudioOutputDevice.addAccess(DeviceAccess("pulse", "default"));
+            newDeviceList.append(defaultAudioOutputDevice);
             return;
         } else {
             pulse->enable(false);
@@ -240,65 +228,79 @@ void DeviceManager::updateDeviceList()
 #endif
 
     QList<QByteArray> knownSoundSystems;
-    knownSoundSystems << "alsa" << "oss";
+    // Whitelist
+    knownSoundSystems << QByteArray("alsa")
+                      << QByteArray("oss")
+                      << QByteArray("jack")
+                      << QByteArray("directx") // Windows
+                      << QByteArray("auhal"); // Mac
     foreach (const QByteArray &soundSystem, knownSoundSystems) {
-        if (audioOutBackends.contains(soundSystem)) {
-            const int deviceCount = libvlc_audio_output_device_count(libvlc, soundSystem);
+        if (!audioOutBackends.contains(soundSystem)) {
+            debug() << "Sound system" << soundSystem << "not supported by libvlc";
+            continue;
+        }
 
-            for (int i = 0; i < deviceCount; i++) {
-                const char *idName = libvlc_audio_output_device_id(libvlc, soundSystem, i);
-                const char *longName = libvlc_audio_output_device_longname(libvlc, soundSystem, i);
+        const int deviceCount = libvlc_audio_output_device_count(libvlc, soundSystem);
 
-                DeviceInfo device(longName, QByteArray() /* no description, sorry */, false);
-                device.addAccess(DeviceAccess(soundSystem, idName));
-                device.setCapabilities(DeviceInfo::AudioOutput);
-                newDeviceList.append(device);
-            }
-            break;
+        for (int i = 0; i < deviceCount; i++) {
+            VString idName(libvlc_audio_output_device_id(libvlc, soundSystem, i));
+            VString longName(libvlc_audio_output_device_longname(libvlc, soundSystem, i));
+
+            debug() << "found device" << soundSystem << idName << longName;
+
+            DeviceInfo device(longName, true);
+            device.addAccess(DeviceAccess(soundSystem, idName));
+            device.setCapabilities(DeviceInfo::AudioOutput);
+            newDeviceList.append(device);
+        }
+
+        // libVLC gives no devices for some sound systems, like OSS
+        if (deviceCount == 0) {
+            debug() << "manually injecting sound system" << soundSystem;
+            DeviceInfo device(QString::fromUtf8(soundSystem), true);
+            device.addAccess(DeviceAccess(soundSystem, ""));
+            device.setCapabilities(DeviceInfo::AudioOutput);
+            newDeviceList.append(device);
         }
     }
 
     /*
      * Compares the list with the devices available at the moment with the last list. If
-     * a new device is seen, a signal is emitted. If a device dissapeared, another signal
-     * is emitted. The devices are only from one category (example audio output devices).
+     * a new device is seen, a signal is emitted. If a device disappeared, another signal
+     * is emitted.
      */
 
-    // New and old device counts
-    int newDeviceCount = newDeviceList.count();
-    int oldDeviceCount = m_devices.count();
-
-    for (int i = 0; i < newDeviceCount; ++i) {
-        int id = deviceId(newDeviceList[i].name());
-        if (id == -1) {
+    // Search for added devices
+    for (int i = 0; i < newDeviceList.count(); ++i) {
+        int id = newDeviceList[i].id();
+        if (!listContainsDevice(m_devices, id)) {
             // This is a new device, add it
             m_devices.append(newDeviceList[i]);
-            id = deviceId(newDeviceList[i].name());
             emit deviceAdded(id);
 
-            debug() << "Added backend device" << newDeviceList[i].name() << "with id" << id;
+            debug() << "Added backend device" << newDeviceList[i].name();
         }
     }
 
-    if (newDeviceCount < oldDeviceCount) {
-        // A device was removed
-        for (int i = oldDeviceCount - 1; i >= 0; --i) {
-            QByteArray name = m_devices[i].name();
-            bool found = false;
-            for (int k = newDeviceCount - 1; k >= 0; --k) {
-                if (name == newDeviceList[k].name()) {
-                    found = true;
-                    break;
-                }
-            }
-            if (!found) {
-                emit deviceRemoved(deviceId(name));
-                m_devices.removeAt(i);
-            }
+    // Search for removed devices
+    for (int i = m_devices.count() - 1; i >= 0; --i) {
+        int id = m_devices[i].id();
+        if (!listContainsDevice(newDeviceList, id)) {
+            emit deviceRemoved(id);
+            m_devices.removeAt(i);
         }
     }
 }
 
+bool DeviceManager::listContainsDevice(const QList<DeviceInfo> &list, int id)
+{
+    foreach (const DeviceInfo &d, list) {
+        if (d.id() == id)
+            return true;
+    }
+    return false;
+}
+
 }
 }
 
