@@ -24,6 +24,7 @@
 
 #include <QtCore/QStringBuilder>
 #include <QtCore/QUrl>
+#include <QDir>
 
 #include <vlc/libvlc_version.h>
 #include <vlc/vlc.h>
@@ -46,11 +47,6 @@ MediaObject::MediaObject(QObject *parent)
     , m_nextSource(MediaSource(QUrl()))
     , m_streamReader(0)
     , m_state(Phonon::StoppedState)
-    // By default, no tick() signal
-    // FIXME: Not implemented yet
-    #ifdef __GNUC__
-    #warning implement tick proper
-    #endif
     , m_tickInterval(0)
     , m_transitionTime(0)
     , m_media(0)
@@ -63,12 +59,13 @@ MediaObject::MediaObject(QObject *parent)
 
     // Player signals.
     connect(m_player, SIGNAL(seekableChanged(bool)), this, SIGNAL(seekableChanged(bool)));
-    connect(m_player, SIGNAL(timeChanged(qint64)), this, SLOT(updateTime(qint64)));
+    connect(m_player, SIGNAL(timeChanged(qint64)), this, SLOT(timeChanged(qint64)));
     connect(m_player, SIGNAL(stateChanged(MediaPlayer::State)), this, SLOT(updateState(MediaPlayer::State)));
     connect(m_player, SIGNAL(hasVideoChanged(bool)), this, SLOT(onHasVideoChanged(bool)));
+    connect(m_player, SIGNAL(bufferChanged(int)), this, SLOT(setBufferStatus(int)));
+    connect(m_player, SIGNAL(timeChanged(qint64)), this, SLOT(timeChanged(qint64)));
 
     // Internal Signals.
-    connect(this, SIGNAL(tickInternal(qint64)), SLOT(tickInternalSlot(qint64)));
     connect(this, SIGNAL(moveToNext()), SLOT(moveToNextSource()));
 
     resetMembers();
@@ -89,8 +86,13 @@ void MediaObject::resetMembers()
     m_prefinishEmitted = false;
     m_aboutToFinishEmitted = false;
 
+    m_lastTick = 0;
+
     m_timesVideoChecked = 0;
 
+    m_buffering = false;
+    m_stateAfterBuffering = ErrorState;
+
     resetMediaController();
 }
 
@@ -112,8 +114,6 @@ void MediaObject::play()
         playInternal();
         break;
     }
-
-    emit playbackCommenced();
 }
 
 void MediaObject::pause()
@@ -159,26 +159,38 @@ void MediaObject::seek(qint64 milliseconds)
         m_aboutToFinishEmitted = false;
 }
 
-void MediaObject::tickInternalSlot(qint64 currentTime)
+void MediaObject::timeChanged(qint64 time)
 {
-    const qint64 totalTime = this->totalTime();
+    const qint64 totalTime = m_totalTime;
 
-    if (m_tickInterval > 0) {
-        // If _tickInternal == 0 means tick() signal is disabled
-        // Default is _tickInternal = 0
-        emit tick(currentTime);
+    switch (m_state) {
+    case PlayingState:
+    case BufferingState:
+    case PausedState:
+        emitTick(time);
+    default:
+        break;
     }
 
-    if (m_state == Phonon::PlayingState) {
-        if (currentTime >= totalTime - m_prefinishMark) {
+    if (m_state == PlayingState || m_state == BufferingState) { // Buffering is concurrent
+        if (time >= totalTime - m_prefinishMark) {
             if (!m_prefinishEmitted) {
                 m_prefinishEmitted = true;
-                emit prefinishMarkReached(totalTime - currentTime);
+                emit prefinishMarkReached(totalTime - time);
             }
         }
-        if (totalTime > -1 && currentTime >= totalTime - ABOUT_TO_FINISH_TIME) {
+        if (totalTime > -1 && time >= totalTime - ABOUT_TO_FINISH_TIME)
             emitAboutToFinish();
-        }
+    }
+}
+
+void MediaObject::emitTick(qint64 time)
+{
+    if (m_tickInterval == 0) // Make sure we do not ever emit ticks when deactivated.\]
+        return;
+    if (time + m_tickInterval >= m_lastTick) {
+        m_lastTick = time;
+        emit tick(time);
     }
 }
 
@@ -213,14 +225,13 @@ qint32 MediaObject::tickInterval() const
     return m_tickInterval;
 }
 
-void MediaObject::setTickInterval(qint32 tickInterval)
+/**
+ * Supports runtime changes.
+ * If the user goes to tick(0) we stop the timer, otherwise we fire it up.
+ */
+void MediaObject::setTickInterval(qint32 interval)
 {
-    m_tickInterval = tickInterval;
-//    if (_tickInterval <= 0) {
-//        _tickTimer->setInterval(50);
-//    } else {
-//        _tickTimer->setInterval(_tickInterval);
-//    }
+    m_tickInterval = interval;
 }
 
 qint64 MediaObject::currentTime() const
@@ -269,6 +280,10 @@ void MediaObject::setSource(const MediaSource &source)
         m_streamReader->unlock();
         delete m_streamReader;
         m_streamReader = 0;
+        // For streamreaders we exchanage the player's seekability with the
+        // reader's so here we change it back.
+        // Note: the reader auto-disconnects due to destruction.
+        connect(m_player, SIGNAL(seekableChanged(bool)), this, SIGNAL(seekableChanged(bool)));
     }
 
     // Reset previous isScreen flag
@@ -276,6 +291,7 @@ void MediaObject::setSource(const MediaSource &source)
 
     m_mediaSource = source;
 
+    QByteArray url;
     switch (source.type()) {
     case MediaSource::Invalid:
         error() << Q_FUNC_INFO << "MediaSource Type is Invalid:" << source.type();
@@ -286,7 +302,13 @@ void MediaObject::setSource(const MediaSource &source)
     case MediaSource::LocalFile:
     case MediaSource::Url:
         debug() << "MediaSource::Url:" << source.url();
-        loadMedia(source.url().toEncoded());
+        if (source.url().scheme().isEmpty()) {
+            url = "file:///";
+            if (source.url().isRelative())
+                url.append(QFile::encodeName(QDir::currentPath()) + '/');
+        }
+        url += source.url().toEncoded();
+        loadMedia(url);
         break;
     case MediaSource::Disc:
         switch (source.discType()) {
@@ -302,6 +324,10 @@ void MediaObject::setSource(const MediaSource &source)
         case Phonon::Vcd:
             loadMedia(m_mediaSource.deviceName());
             break;
+#if (PHONON_VERSION >= PHONON_VERSION_CHECK(4, 7, 0))
+        case Phonon::BluRay:
+            loadMedia(QLatin1Literal("bluray://") % m_mediaSource.deviceName());
+#endif
         }
         break;
     case MediaSource::CaptureDevice: {
@@ -350,7 +376,15 @@ void MediaObject::setSource(const MediaSource &source)
         break;
     }
     case MediaSource::Stream:
-        m_streamReader = new StreamReader(m_mediaSource, this);
+        m_streamReader = new StreamReader(this);
+        // LibVLC refuses to emit seekability as it does a try-and-seek approach
+        // to work around this we exchange the player's seekability signal
+        // for the readers
+        // https://bugs.kde.org/show_bug.cgi?id=293012
+        connect(m_streamReader, SIGNAL(streamSeekableChanged(bool)), this, SIGNAL(seekableChanged(bool)));
+        disconnect(m_player, SIGNAL(seekableChanged(bool)), this, SIGNAL(seekableChanged(bool)));
+        // Only connect now to avoid seekability detection before we are connected.
+        m_streamReader->connectToSource(source);
         loadMedia(QByteArray("imem://"));
         break;
     }
@@ -402,8 +436,6 @@ void MediaObject::emitAboutToFinish()
 void MediaObject::changeState(Phonon::State newState)
 {
     DEBUG_BLOCK;
-    debug() << m_state << "-->" << newState;
-
     if (newState == m_state) {
         // State not changed
         return;
@@ -412,6 +444,9 @@ void MediaObject::changeState(Phonon::State newState)
         debug() << Q_FUNC_INFO << "no-op gapless item awaiting in queue - " << m_nextSource.type() ;
         return;
     }
+
+    debug() << m_state << "-->" << newState;
+
 #ifdef __GNUC__
 #warning do we actually need m_seekpoint? if a consumer seeks before playing state that is their problem?!
 #endif
@@ -442,7 +477,7 @@ void MediaObject::moveToNextSource()
     m_nextSource = MediaSource(QUrl());
 }
 
-bool MediaObject::checkGaplessWaiting()
+inline bool MediaObject::checkGaplessWaiting()
 {
     return m_nextSource.type() != MediaSource::Invalid && m_nextSource.type() != MediaSource::Empty;
 }
@@ -459,9 +494,9 @@ inline void MediaObject::unloadMedia()
 void MediaObject::playInternal()
 {
     DEBUG_BLOCK;
-    unloadMedia();
 
-    m_totalTime = -1;
+    unloadMedia();
+    resetMembers();
 
     // Create a media with the given MRL
     m_media = new Media(m_mrl, this);
@@ -519,14 +554,18 @@ bool MediaObject::hasVideo() const
 
 bool MediaObject::isSeekable() const
 {
+    if (m_streamReader)
+        return m_streamReader->streamSeekable();
     return m_player->isSeekable();
 }
 
 void MediaObject::updateDuration(qint64 newDuration)
 {
-#ifdef __GNUC__
-#warning duration signal can just be forwarded, we have no gain from caching this
-#endif
+    // This here cache is needed because we need to provide -1 as totalTime()
+    // for as long as we do not get a proper update through this slot.
+    // VLC reports -1 with no media but 0 if it does not know the duration, so
+    // apps that assume 0 = unkown get screwed if they query too early.
+    // http://bugs.tomahawk-player.org/browse/TWK-1029
     m_totalTime = newDuration;
     emit totalTimeChanged(m_totalTime);
 }
@@ -571,6 +610,7 @@ void MediaObject::updateState(MediaPlayer::State state)
 {
     DEBUG_BLOCK;
     debug() << state;
+
     switch (state) {
     case MediaPlayer::NoState:
         changeState(LoadingState);
@@ -580,7 +620,6 @@ void MediaObject::updateState(MediaPlayer::State state)
         break;
     case MediaPlayer::BufferingState:
         changeState(BufferingState);
-        emit bufferStatus(m_player->bufferCache());
         break;
     case MediaPlayer::PlayingState:
         changeState(PlayingState);
@@ -589,51 +628,58 @@ void MediaObject::updateState(MediaPlayer::State state)
         changeState(PausedState);
         break;
     case MediaPlayer::StoppedState:
-        resetMembers();
         changeState(StoppedState);
         break;
     case MediaPlayer::EndedState:
-        resetMembers();
         emitAboutToFinish();
         emit finished();
         changeState(StoppedState);
         break;
     case MediaPlayer::ErrorState:
-        resetMembers();
+        debug() << errorString();
         emitAboutToFinish();
         emit finished();
         changeState(ErrorState);
         break;
     }
-}
 
-void MediaObject::updateTime(qint64 time)
-{
-    // Only VLC >= 2.0 implements a signal on vout appearance.
-    // On older versions we simply probe for it...
-#if (LIBVLC_VERSION_INT < LIBVLC_VERSION(2, 0, 0, 0))
-    // Check 10 times for a video, then give up.
-    if (!m_hasVideo && ++m_timesVideoChecked < 11) {
-        debug() << "Looking for Video";
-        onHasVideoChanged(m_player->hasVideoOutput());
+    if (m_buffering) {
+        switch (state) {
+        case MediaPlayer::BufferingState:
+            break;
+        case MediaPlayer::PlayingState:
+            debug() << "Restoring buffering state after state change to Playing";
+            changeState(BufferingState);
+            m_stateAfterBuffering = PlayingState;
+            break;
+        case MediaPlayer::PausedState:
+            debug() << "Restoring buffering state after state change to Paused";
+            changeState(BufferingState);
+            m_stateAfterBuffering = PausedState;
+            break;
+        default:
+            debug() << "Buffering aborted!";
+            m_buffering = false;
+            break;
+        }
     }
-#endif
 
-    emit tickInternal(time);
+    return;
 }
 
 void MediaObject::onHasVideoChanged(bool hasVideo)
 {
+    DEBUG_BLOCK;
     if (m_hasVideo != hasVideo) {
         m_hasVideo = hasVideo;
         emit hasVideoChanged(m_hasVideo);
-    }
+    } else
+        // We can simply return if we are have the appropriate caching already.
+        // Otherwise we'd do pointless rescans of mediacontroller stuff.
+        // MC and MO are force-reset on media changes anyway.
+        return;
 
     if (hasVideo) {
-        debug() << "HASVIDEO";
-#ifdef __GNUC__
-#warning a bit inperformant and stuff
-#endif
         refreshAudioChannels();
         refreshSubtitles();
 
@@ -646,6 +692,34 @@ void MediaObject::onHasVideoChanged(bool hasVideo)
     }
 }
 
+void MediaObject::setBufferStatus(int percent)
+{
+    // VLC does not have a buffering state (surprise!) but instead only sends the
+    // event (surprise!). Hence we need to simulate the state change.
+    // Problem with BufferingState is that it is actually concurent to Playing or Paused
+    // meaning that while you are buffering you can also pause, thus triggering
+    // a state change to paused. To handle this we let updateState change the
+    // state accordingly (as we need to allow the UI to update itself, and
+    // immediately after that we change back to buffering again.
+    // This loop can only be interrupted by a state change to !Playing & !Paused
+    // or by reaching a 100 % buffer caching (i.e. full cache).
+
+    m_buffering = true;
+    if (m_state != BufferingState) {
+        m_stateAfterBuffering = m_state;
+        changeState(BufferingState);
+    }
+
+    emit bufferStatus(percent);
+
+    // Transit to actual state only after emission so the signal is still
+    // delivered while in BufferingState.
+    if (percent >= 100) { // http://trac.videolan.org/vlc/ticket/5277
+        m_buffering = false;
+        changeState(m_stateAfterBuffering);
+    }
+}
+
 qint64 MediaObject::totalTime() const
 {
     return m_totalTime;
