@@ -23,6 +23,7 @@
 
 #include <vlc/vlc.h>
 
+class QImage;
 class QString;
 
 namespace Phonon {
@@ -49,10 +50,13 @@ public:
     ~MediaPlayer();
 
     inline libvlc_media_player_t *libvlc_media_player() const { return m_player; }
-    inline float bufferCache() const { return m_bufferCache; }
+    inline operator libvlc_media_player_t *() const { return m_player; }
 
     void setMedia(Media *media);
 
+    void setVideoCallbacks();
+    void setVideoFormatCallbacks();
+
     void setNsObject(void *drawable) { libvlc_media_player_set_nsobject(m_player, drawable); }
     void setXWindow(quint32 drawable) { libvlc_media_player_set_xwindow(m_player, drawable); }
     void setHwnd(void *drawable) { libvlc_media_player_set_hwnd(m_player, drawable); }
@@ -64,6 +68,7 @@ public:
     void togglePause();
     void stop();
 
+    qint64 length() const;
     qint64 time() const;
     void setTime(qint64 newTime);
 
@@ -116,6 +121,9 @@ public:
 
     void setChapter(int chapter);
 
+    /** Reentrant, through libvlc */
+    QImage snapshot() const;
+
     // Audio
     /// Get current audio volume.
     /// \return the software volume in percents (0 = mute, 100 = nominal / 0dB)
@@ -151,6 +159,7 @@ signals:
     void seekableChanged(bool seekable);
     void stateChanged(MediaPlayer::State state);
     void timeChanged(qint64 time);
+    void bufferChanged(int percent);
 
     /** Emitted when the vout availability has changed */
     void hasVideoChanged(bool hasVideo);
@@ -161,8 +170,6 @@ private:
     Media *m_media;
 
     libvlc_media_player_t *m_player;
-
-    float m_bufferCache;
 };
 
 QDebug operator<<(QDebug dbg, const MediaPlayer::State &s);
